/* Automatically generated by create_config - do not modify */
#define HOST_I386 1
#define CONFIG_WIN32 1
#define CONFIG_FILEVERSION 2,2,1,0
#define CONFIG_PRODUCTVERSION 2,2,1,0
#define CONFIG_CPUID_H 1
